/*
    scx-xx: Samsung SCX-4200 scanner driver
    Copyright (C) 2007-2008 Vladimir Koutny

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include "scx-xx.h"

#define ERROR(msg...) { fprintf(stderr, "%ld: ", time(NULL)); fprintf(stderr, msg); }

#define SET_4B(ptr, value) { \
    *(ptr+0)=(value >> 24) & 0xFF; \
    *(ptr+1)=(value >> 16) & 0xFF; \
    *(ptr+2)=(value >> 8) & 0xFF; \
    *(ptr+3)=(value >> 0) & 0xFF; \
    }

#define SET_2B(ptr, value) { \
    *(ptr+0)=(value >> 8) & 0xFF; \
    *(ptr+1)=(value >> 0) & 0xFF; \
    }

#define SET_1B(ptr, value)  *(ptr+0) = (value & 0xFF);

#define GET_4B(ptr) ( \
    (*(ptr+0) << 24) | \
    (*(ptr+1) << 16) | \
    (*(ptr+2) << 8) | \
    (*(ptr+3) << 0) )

#define GET_2B(ptr) ( \
    (*(ptr+0) << 8) | \
    (*(ptr+1) << 0) )

unsigned char sams_cmd_DeviceInfo[]  = {4, 0x1B, 0xA8, 0x12, 0x00};
unsigned char sams_cmd_StartConfig[] = {4, 0x1B, 0xA8, 0x16, 0x00};
unsigned char sams_cmd_Start[]       = {4, 0x1B, 0xA8, 0x31, 0x00};
unsigned char sams_cmd_Stop[]        = {4, 0x1B, 0xA8, 0x06, 0x00};
unsigned char sams_cmd_Status[]      = {4, 0x1B, 0xA8, 0x28, 0x00};
unsigned char sams_cmd_SendData[]    = {4, 0x1B, 0xA8, 0x29, 0x00};
unsigned char sams_cmd_Config[]      = {25, 
    0x1B, 0xA8, 0x24, 0x13,
    0x30, 
    0x00, 0x00, 0x00, 0x00,   // width
    0x00, 0x00, 0x00, 0x00,   // height
    0x00, 0x00,               // resolution
    0x00, 0x00,               // left
    0x00, 0x00,               // top
    0x03,                     // color mode
    0x00, 0x00, 0x02, 0x40, 0x00
    };

char *sams_buf_ev_names[] = {"None", "Ok", "Data", "DataLast", "NoData", "DeviceInfo", "Error", "Unknown"};
char *sams_buf_st_names[] = {"Idle", "StartConfig", "DoConfig", "PreScan", "Scanning", "PostScan", "Done"};

char *sams_get_st_name(enum SamsState st) {
    return sams_buf_st_names[st];
}
char *sams_get_ev_name(enum SamsEvent ev) {
    return sams_buf_ev_names[ev];
}

void sams_dump(char *prefix, unsigned char *buf, int len, int ascii) 
{
    int i;
    if (prefix) fprintf(stderr, "%s: ", prefix);
    
    for (i=0; i<len; i++) {
        if ((buf[i] < 32) || (buf[i] > 127) || (!ascii)) 
	    fprintf(stderr, "[%02X]", buf[i]);
	else
	    fprintf(stderr, "%c", buf[i]);
    }
    fprintf(stderr, "\n");
}


int sams_cmd(int fd, enum SamsCommand cmd, int *args)
{
    unsigned char buf[64];
    unsigned char *q = NULL;
    int res;
    unsigned int left, top, width, height;
    
    switch (cmd) {
	case cmd_DeviceInfo:  if (!q) q=sams_cmd_DeviceInfo;
	case cmd_StartConfig: if (!q) q=sams_cmd_StartConfig;
	case cmd_Start:       if (!q) q=sams_cmd_Start;
	case cmd_Stop:        if (!q) q=sams_cmd_Stop;
	case cmd_Status:      if (!q) q=sams_cmd_Status;
	case cmd_SendData:    if (!q) q=sams_cmd_SendData;
	    break;
    
	case cmd_Config:
	    memcpy(buf, sams_cmd_Config, sams_cmd_Config[0]+1);
	    
	    // fill in scan parameters
	    left = (int)(1.0*args[2]/25.4*100);
	    top  = (int)(1.0*args[3]/25.4*100);
	    width = (int)(1.0*args[4]/25.4*1200-47);
	    height = (int)(1.0*args[5]/25.4*1200-47);
	    
	    SET_4B(buf+1+SAMS_CONF_WIDTH,  width);
	    SET_4B(buf+1+SAMS_CONF_HEIGHT, height);
	    SET_2B(buf+1+SAMS_CONF_RES,    args[0]);
	    SET_2B(buf+1+SAMS_CONF_LEFT,   ((left/100)<<8 | (left%100))); 
	    SET_2B(buf+1+SAMS_CONF_TOP,    ((top/100)<<8 | (top%100))); 
	    SET_1B(buf+1+SAMS_CONF_COLOR,  args[1]);
	    
	    ERROR("CONFIG: %dx%d @ %d,%d\n", width, height, left, top);
	    // sams_dump("CONFIG", buf+1, buf[0], 0);
	    q=buf;
	    break;
    }
    
    if (q) {
	res = 0;
	// sams_dump("SEND", q+1, q[0], 0);
	// if (cmd!=cmd_Config) 
	res = write(fd, q+1, q[0]);
    } else res = -1;
    
    return res;
}

enum SamsEvent sams_wait_event(int fd, unsigned char *buf, int buflen, unsigned char **bufptr, int *msglen)
{
    fd_set fds;
    struct timeval timeout;
    int res, datalen;
    enum SamsEvent event = ev_None;

    *msglen = 0;
    
    // check if we have an event in the buffer
    datalen = *bufptr - buf;
    if (datalen >= 4) {
	// sams_dump("RECV", buf, 16, 0);
	if (buf[0] != 0xA8) {
	    // error, message header sould start here
	    ERROR("Invalid event header!\n");
	    *bufptr = buf;
	}
	else {
	    int len = buf[2];
	    if (datalen >= len+3) {
		// we have enough data received -> process it
		
		switch (buf[3] & 0xF0) {
		    case 0x00: event=ev_Ok; break;
		    case 0x10: event=ev_DeviceInfo; break;
		    case 0x20: event=ev_Error; break;
		    case 0x80: 
			switch (buf[1]) {
			    case 0x08: event=ev_NoData; break;
			    case 0x00: if (buf[3]&0x01) event=ev_DataLast; else event=ev_Data; break;
			}
			break;
			
		    default: event=ev_Unknown;
		}
		*msglen = len+3;
	    }
	}
    }
    
    if (event != ev_None) 
	return event;
    
    
    // wait for some data
    FD_ZERO(&fds);
    FD_SET(fd, &fds);
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;    
    
    res = select(fd+1, &fds, NULL, NULL, &timeout);
    if (res < 0) {	// Error, wait a bit to prevent busy waiting
	ERROR("SELECT error: %d\n", errno);
	usleep(10000);
    } 
    else if (res > 0) {
	if (FD_ISSET(fd, &fds)) {
	    int spaceleft = buflen - (*bufptr - buf);
	    res = read(fd, *bufptr, spaceleft);
	    // ERROR("reading %d bytes: buf=%p, ptr=%p, left=%d\n", res, buf, *bufptr, spaceleft);
	    if (res > 0)
		*bufptr += res;
	}
    }
    
    return ev_None;
}

void sams_consume_event(unsigned char *buf, int buflen, unsigned char **bufptr, int msglen)
{
    int len = *bufptr - buf;
    
    if (len < msglen) *bufptr = buf;
    else {
	unsigned char *q = buf+msglen;
	unsigned char *r = buf;
	
	while (q < *bufptr) {
	    *r = *q;
	    r++, q++;	    
	}
	*bufptr -= msglen;
    }
}

enum SamsState sams_state_idle(int fd, enum SamsEvent ev, unsigned char *buf, int msglen)
{
    if (ev == ev_DeviceInfo) {
	sams_dump("Device info", buf, 20, 1);
	
	sams_cmd(fd, cmd_StartConfig, NULL);

	return st_StartConfig;
    }
    
    return st_Idle;
}

enum SamsState sams_state_start_config(int fd, enum SamsEvent ev, unsigned char *buf, int msglen, struct SamsConfiguration *config)
{
    if (ev == ev_Ok) {
	int data[6];
	
	data[0] = config->resolution;
	data[1] = config->colormode;
	data[2] = config->left;
	data[3] = config->top;
	data[4] = config->width;
	data[5] = config->height;

	sams_cmd(fd, cmd_Config, data);

	return st_DoConfig;
    }


    return st_StartConfig;
}

enum SamsState sams_state_do_config(int fd, enum SamsEvent ev, unsigned char *buf, int msglen)
{
    if (ev == ev_Ok) {
	ERROR("* Starting scanning..\n");
	sams_cmd(fd, cmd_Start, NULL);
	return st_PreScan;
    }
    
    return st_DoConfig;
}

enum SamsState sams_state_pre_scan(int fd, enum SamsEvent ev, unsigned char *buf, int msglen)
{
    if (ev == ev_Ok) {
	// received confirmation 
	return st_Scanning;
    }

    return st_PreScan;
}

enum SamsState sams_state_scanning(int fd, enum SamsEvent ev, unsigned char *buf, 
    int msglen, unsigned char **bufptr, enum SamsColors color, int *width, int *height)
{
    static int dont_send = 0;
    
    if (ev == ev_None) {
	// send status query
	if (!dont_send)
	    sams_cmd(fd, cmd_Status, NULL);
	dont_send = 1;
	return st_Scanning;
    }
    dont_send = 0;
    
    if (ev == ev_Ok) {
	// finished, send stop command
	sams_cmd(fd, cmd_Stop, NULL);
	return st_PostScan;
    }
    
    if (ev == ev_NoData) {
	ERROR("* no data yet\n");
	return st_Scanning;
    };
    
    if ((ev == ev_Data) || (ev == ev_DataLast)) {
	int datalen, res, left, done;
	unsigned char *databuf;
	
	datalen = GET_4B(buf+SAMS_DATA_SIZE);
	*width = GET_2B(buf+SAMS_DATA_WIDTH);
	*height = GET_2B(buf+SAMS_DATA_HEIGHT);
	
	// fetch all data
	databuf = (unsigned char *)malloc(datalen+1);
	// ERROR("* got %d bytes of data, will save to %p\n", datalen, databuf);

	sams_cmd(fd, cmd_SendData, NULL);

	done = 0;
	left = datalen;
	while (left>0) {
	    int size = 8192;
	    if (left<size) size = left;
	    // ERROR("SCAN DATA: done %d, left %d, size %d\n", done, left, size);
	    
	    res = read(fd, databuf+done, size);
	    done+=res;
	    left-=res;
	}
	ERROR("SCAN DATA: %d bytes, %dx%d\n", done, *width, *height);
	if (color == col_Truecolor) {
	    // merge 3 lines into 1 rgb line
	    unsigned char merge[65536];
	    int x,y,c;
	    unsigned char *q, *r=databuf;
	    
	    for (y=0; y<*height; y++) {
		for (c=0; c<3; c++) {
		    q = &merge[c];
		    for (x=0; x<*width; x++) {
			*q = *r;
			q+=3;
			r++;
		    }
		}
		fwrite(merge, 1, 3*(*width), stdout);
	    }
	}
	else fwrite(databuf, 1, done-16, stdout);
	free(databuf);
	
	if (ev == ev_DataLast) {
	    sams_cmd(fd, cmd_Stop, NULL);
	    return st_PostScan;
	}
	
	return st_Scanning;
    };
    
    return st_Scanning;
}

enum SamsState sams_state_post_scan(int fd, enum SamsEvent ev, unsigned char *buf, int msglen)
{
    if (ev == ev_Ok) {
	return st_Done;
    }

    return st_PostScan;
}


int main(int argc, char **argv) 
{
    
    int fd;
    enum SamsState st, oldstate=st_Idle;
    unsigned char buf[1024];
    unsigned char *bufptr = buf;
    int msglen, i;
    int total_width, total_height;

    struct SamsConfiguration config;


    // default config
    config.width = 130;
    config.height = 100;
    config.left = 30;
    config.top = 50;
    config.resolution = res_100dpi;
    config.colormode = col_Truecolor;

    for (i=1; i<argc; i++) {
	if (!strcmp(argv[i],"-l")) sscanf(argv[++i], "%d", &config.left);
	if (!strcmp(argv[i],"-t")) sscanf(argv[++i], "%d", &config.top);
	if (!strcmp(argv[i],"-w")) sscanf(argv[++i], "%d", &config.width);
	if (!strcmp(argv[i],"-h")) sscanf(argv[++i], "%d", &config.height);

	if (!strcmp(argv[i],"-r")) {
	    int res = 100;
	    sscanf(argv[++i], "%d", &res);
	    switch (res) {
		case 75:  config.resolution = res_75dpi; break;
		case 100: config.resolution = res_100dpi; break;
		case 150: config.resolution = res_150dpi; break;
		case 200: config.resolution = res_200dpi; break;
		case 300: config.resolution = res_300dpi; break;
		case 600: config.resolution = res_600dpi; break;
	    }
	}

	if (!strcmp(argv[i],"-c")) {
	    i++;
	    if (!strcmp(argv[i],"color")) config.colormode = col_Truecolor;
	    if (!strcmp(argv[i],"gray"))  config.colormode = col_Grayscale;
	    if (!strcmp(argv[i],"lineart")) config.colormode = col_Lineart;
	    if (!strcmp(argv[i],"halftone")) config.colormode = col_Halftone;
	}
    }

    ERROR("Starting scan: %dx%dmm (@%d,%dmm)\n", config.width, config.height, config.left, config.top);    
    fd = open("/dev/usb/lp0", O_RDWR | O_EXCL);

    st = st_Idle;
    total_width = 0;
    total_height = 0;
    
    sams_cmd(fd, cmd_DeviceInfo, NULL);

    while (st != st_Done) {
	int w,h;
	enum SamsEvent ev = sams_wait_event(fd, buf, sizeof(buf), &bufptr, &msglen);
	
	if (ev != ev_None)
	    fprintf(stderr, "[st:%s, ev:%s]\n", sams_get_st_name(st), sams_get_ev_name(ev));
	
	switch(st) {
	    case st_Idle:
		st = sams_state_idle(fd, ev, buf, msglen);
		break;

	    case st_StartConfig:
		st = sams_state_start_config(fd, ev, buf, msglen, &config);
		break;

	    case st_DoConfig:
		st = sams_state_do_config(fd, ev, buf, msglen);
		break;

	    case st_PreScan:
		st = sams_state_pre_scan(fd, ev, buf, msglen);
		break;
	
	    case st_Scanning:
		w=0; h=0;
		st = sams_state_scanning(fd, ev, buf, msglen, &bufptr, config.colormode, &w, &h);
		total_width = (w>0 ? w:total_width);
		total_height += h;
		break;
		
	    case st_PostScan:
		st = sams_state_post_scan(fd, ev, buf, msglen);
		break;
		
	    case st_Done:
		break;
	}
	
	// delete processed event from input buffer
	if (ev != ev_None)
	    sams_consume_event(buf, sizeof(buf), &bufptr, msglen);
	    
	if (oldstate != st)
	    ERROR(" -> [st:%s]\n", sams_get_st_name(st));
	oldstate = st;
    }

    ERROR("Finished, total pixel size is %d x %d\n", total_width, total_height);
    
    close(fd);
    return 0;
}
